#!/bin/bash

# cryptest.sh - written and placed in public domain by Jeffrey Walton and Uri Blumenthal.
#               Copyright assigned to Crypto++ project.

# This is a test script that can be used on some Linux/Unix/Apple machines
# to automate building the library and running the self test with various
# combinations of flags, options, and conditions.

# Everything is tee'd into cryptest-result.txt. You should be able to
# use `egrep "(error|FAILED)" cryptest-result.txt` to quickly find
# errors and failures.

# Set to suite your taste
TEST_RESULTS=cryptest-result.txt

# Respect user's preferred flags, but filter the stuff we expliclty test
#if [ ! -z "CXXFLAGS" ]; then
#	ADD_CXXFLAGS=$(echo "$CXXFLAGS" | sed 's/\(-DDEBUG\|-DNDEBUG\|-O[0-9]\|-Os\|-Og\|-fsanitize=address\|-fsanitize=undefined\|-DDCRYPTOPP_NO_UNALIGNED_DATA_ACCESS\|-DDCRYPTOPP_NO_UNALIGNED_DATA_ACCESS\|-DDCRYPTOPP_NO_BACKWARDS_COMPATIBILITY_562\)//g')
#else\
#	ADD_CXXFLAGS=""
#fi

# I can't seem to get the expression to work in sed on Apple. It returns the original CXXFLAGS.
#   If you want to test with additional flags, the put them in ADD_CXXFLAGS below.
# ADD_CXXFLAGS="-mrdrnd -mrdseed"
ADD_CXXFLAGS=""

cpp -dM -std=c++11 < /dev/null > /dev/null 2>&1
if [ "$?" -eq "0" ]; then
	HAVE_CXX11=1
else
	HAVE_CXX11=0
fi

# Set to 0 if you don't have UBsan
# Also see https://unix.stackexchange.com/questions/3514/how-to-grep-standard-error-stream-stderr
cpp -dM -fsanitize=undefined < /dev/null > /dev/null 2>&1
if [ "$?" -eq "0" ]; then
	HAVE_UBSAN=1
else
	HAVE_UBSAN=0
fi

# Set to 0 if you don't have Asan
cpp -dM -fsanitize=address < /dev/null > /dev/null 2>&1
if [ "$?" -eq "0" ]; then
	HAVE_ASAN=1
else
	HAVE_ASAN=0
fi

# Set to 0 if you don't have Valgrind. Valgrind tests take a long time...
HAVE_VALGRIND=$(which valgrind > /dev/null 2>&1 | grep -i -c valgrind)

IS_DARWIN=$(uname -s | grep -i -c darwin)

echo "HAVE_CXX11: $HAVE_CXX11"
echo "HAVE_ASAN: $HAVE_ASAN"
echo "HAVE_UBSAN: $HAVE_UBSAN"
echo "HAVE_VALGRIND: $HAVE_VALGRIND"
echo "IS_DARWIN: $IS_DARWIN"
echo "User CXXFLAGS: $CXXFLAGS"
echo "Retained CXXFLAGS: $ADD_CXXFLAGS"
echo

# Remove previous test results
rm -f "$TEST_RESULTS" > /dev/null 2>&1
touch "$TEST_RESULTS"

TEST_BEGIN=$(date)

# Basic Debug build
echo "************************************" | tee -a "$TEST_RESULTS"
echo "Debug build, no user flags" | tee -a "$TEST_RESULTS"
echo

unset CXXFLAGS
make clean > /dev/null 2>&1
export CXXFLAGS="-DDEBUG -g2 -O2"
make static dynamic cryptest.exe
./cryptest.exe v 2>&1 | tee -a "$TEST_RESULTS"
./cryptest.exe tv all 2>&1 | tee -a "$TEST_RESULTS"

# Basic Release build
echo "************************************" | tee -a "$TEST_RESULTS"
echo "Release build, no user flags" | tee -a "$TEST_RESULTS"
echo

unset CXXFLAGS
make clean > /dev/null 2>&1
export CXXFLAGS="-DNDEBUG -g2 -O2"
make static dynamic cryptest.exe
./cryptest.exe v 2>&1 | tee -a "$TEST_RESULTS"
./cryptest.exe tv all 2>&1 | tee -a "$TEST_RESULTS"

# C++03 Debug build
echo "************************************" | tee -a "$TEST_RESULTS"
echo "Debug build, C++03" | tee -a "$TEST_RESULTS"
echo

unset CXXFLAGS
make clean > /dev/null 2>&1
export CXXFLAGS="-DDEBUG -g2 -O2 -std=c++03 $ADD_CXXFLAGS"
make static dynamic cryptest.exe
./cryptest.exe v 2>&1 | tee -a "$TEST_RESULTS"
./cryptest.exe tv all 2>&1 | tee -a "$TEST_RESULTS"

# C++03 Release build
echo "************************************" | tee -a "$TEST_RESULTS"
echo "Release build, C++03" | tee -a "$TEST_RESULTS"
echo

unset CXXFLAGS
make clean > /dev/null 2>&1
export CXXFLAGS="-DNDEBUG -g2 -O2 -std=c++03 $ADD_CXXFLAGS"
make static dynamic cryptest.exe
./cryptest.exe v 2>&1 | tee -a "$TEST_RESULTS"
./cryptest.exe tv all 2>&1 | tee -a "$TEST_RESULTS"

# C++11 Debug build
if [ "$HAVE_CXX11" -ne "0" ]; then
	echo "************************************" | tee -a "$TEST_RESULTS"
	echo "Debug build, C++11" | tee -a "$TEST_RESULTS"
	echo

	unset CXXFLAGS
	make clean > /dev/null 2>&1
	export CXXFLAGS="-DDEBUG -g2 -O2 -std=c++11 $ADD_CXXFLAGS"
	make static dynamic cryptest.exe 2>&1 | tee -a "$TEST_RESULTS"
	./cryptest.exe v 2>&1 | tee -a "$TEST_RESULTS"
	./cryptest.exe tv all 2>&1 | tee -a "$TEST_RESULTS"
fi

# C++11 Release build
if [ "$HAVE_CXX11" -ne "0" ]; then
	echo "************************************" | tee -a "$TEST_RESULTS"
	echo "Release build, C++11" | tee -a "$TEST_RESULTS"
	echo

	unset CXXFLAGS
	make clean > /dev/null 2>&1
	export CXXFLAGS="-DNDEBUG -g2 -O2 -std=c++11 $ADD_CXXFLAGS"
	make static dynamic cryptest.exe
	./cryptest.exe v 2>&1 | tee -a "$TEST_RESULTS"
	./cryptest.exe tv all 2>&1 | tee -a "$TEST_RESULTS"
fi

# Release build, with all backwards compatibility.
echo "************************************" | tee -a "$TEST_RESULTS"
echo "Release build, with all compatibility" | tee -a "$TEST_RESULTS"
echo

unset CXXFLAGS
make clean > /dev/null 2>&1
export CXXFLAGS="-DNDEBUG -g2 -O3 -DCRYPTOPP_MAINTAIN_BACKWARDS_COMPATIBILITY $ADD_CXXFLAGS"
make static dynamic cryptest.exe
./cryptest.exe v 2>&1 | tee -a "$TEST_RESULTS"
./cryptest.exe tv all 2>&1 | tee -a "$TEST_RESULTS"

# Release build, no unaligned data access
#  This test will not be needed in Crypto++ 5.7 and above
echo "************************************" | tee -a "$TEST_RESULTS"
echo "Release build, with no unaligned data accesses" | tee -a "$TEST_RESULTS"
echo

unset CXXFLAGS
make clean > /dev/null 2>&1
export CXXFLAGS="-DNDEBUG -g2 -O3 -DCRYPTOPP_NO_UNALIGNED_DATA_ACCESS $ADD_CXXFLAGS"
make static dynamic cryptest.exe
./cryptest.exe v 2>&1 | tee -a "$TEST_RESULTS"
./cryptest.exe tv all 2>&1 | tee -a "$TEST_RESULTS"

# Release build, init_priority
#  This test will not be needed in Crypto++ 5.7 and above
echo "************************************" | tee -a "$TEST_RESULTS"
echo "Release build, with init_priority" | tee -a "$TEST_RESULTS"
echo

unset CXXFLAGS
make clean > /dev/null 2>&1
export CXXFLAGS="-DNDEBUG -g2 -O3 -DCRYPTOPP_INIT_PRIORITY=250 $ADD_CXXFLAGS"
make static dynamic cryptest.exe
./cryptest.exe v 2>&1 | tee -a "$TEST_RESULTS"
./cryptest.exe tv all 2>&1 | tee -a "$TEST_RESULTS"

# Release build, no backwards compatibility with Crypto++ 5.6.2.
#  This test will not be needed in Crypto++ 5.7 and above
echo "************************************" | tee -a "$TEST_RESULTS"
echo "Release build, without 5.6.2 compatibility" | tee -a "$TEST_RESULTS"
echo

unset CXXFLAGS
make clean > /dev/null 2>&1
export CXXFLAGS="-DNDEBUG -g2 -O3 -DCRYPTOPP_NO_BACKWARDS_COMPATIBILITY_562 $ADD_CXXFLAGS"
make static dynamic cryptest.exe
./cryptest.exe v 2>&1 | tee -a "$TEST_RESULTS"
./cryptest.exe tv all 2>&1 | tee -a "$TEST_RESULTS"

# Release build at -O3
echo "************************************" | tee -a "$TEST_RESULTS"
echo "Release build, -O3 optimizations" | tee -a "$TEST_RESULTS"
echo

unset CXXFLAGS
make clean > /dev/null 2>&1
export CXXFLAGS="-DNDEBUG -g2 -O3 $ADD_CXXFLAGS"
make static dynamic cryptest.exe
./cryptest.exe v 2>&1 | tee -a "$TEST_RESULTS"
./cryptest.exe tv all 2>&1 | tee -a "$TEST_RESULTS"

# Release build at -Os
echo "************************************" | tee -a "$TEST_RESULTS"
echo "Release build, -Os optimizations" | tee -a "$TEST_RESULTS"
echo

unset CXXFLAGS
make clean > /dev/null 2>&1
export CXXFLAGS="-DNDEBUG -g2 -Os $ADD_CXXFLAGS"
make static dynamic cryptest.exe
./cryptest.exe v 2>&1 | tee -a "$TEST_RESULTS"
./cryptest.exe tv all 2>&1 | tee -a "$TEST_RESULTS"

# Debug build, UBSan, C++03
if [ "$HAVE_UBSAN" -ne "0" ]; then
	echo "************************************" | tee -a "$TEST_RESULTS"
	echo "Debug build, C++03, Undefined Behavior sanitizer" | tee -a "$TEST_RESULTS"
	echo
	
	unset CXXFLAGS
	make clean > /dev/null 2>&1
	export CXXFLAGS="-DDEBUG -g2 -O2 -std=c++03 $ADD_CXXFLAGS"
	make ubsan
	./cryptest.exe v 2>&1 | tee -a "$TEST_RESULTS"
	./cryptest.exe tv all 2>&1 | tee -a "$TEST_RESULTS"
fi

# Release build, UBSan, C++03
if [ "$HAVE_UBSAN" -ne "0" ]; then
	echo "************************************" | tee -a "$TEST_RESULTS"
	echo "Release build, C++03, Undefined Behavior sanitizer" | tee -a "$TEST_RESULTS"
	echo
	
	unset CXXFLAGS
	make clean > /dev/null 2>&1
	export CXXFLAGS="-DNDEBUG -g2 -O2 -std=c++03 $ADD_CXXFLAGS"
	make ubsan
	./cryptest.exe v 2>&1 | tee -a "$TEST_RESULTS"
	./cryptest.exe tv all 2>&1 | tee -a "$TEST_RESULTS"
fi

# Debug build, Asan, C++03
if [ "$HAVE_ASAN" -ne "0" ]; then
	echo "************************************" | tee -a "$TEST_RESULTS"
	echo "Debug build, C++03, Address sanitizer" | tee -a "$TEST_RESULTS"
	echo
	
	unset CXXFLAGS
	make clean > /dev/null 2>&1
	export CXXFLAGS="-DDEBUG -g2 -O2 -std=c++03 $ADD_CXXFLAGS"
	make asan
	./cryptest.exe v 2>&1 | tee -a "$TEST_RESULTS"
	./cryptest.exe tv all 2>&1 | tee -a "$TEST_RESULTS"
fi

# Release build, Asan, C++03
if [ "$HAVE_ASAN" -ne "0" ]; then
	echo "************************************" | tee -a "$TEST_RESULTS"
	echo "Release build, C++03, Address sanitizer" | tee -a "$TEST_RESULTS"
	echo
	
	unset CXXFLAGS
	make clean > /dev/null 2>&1
	export CXXFLAGS="-DNDEBUG -g2 -O2 -std=c++03 $ADD_CXXFLAGS"
	make asan
	./cryptest.exe v 2>&1 | tee -a "$TEST_RESULTS"
	./cryptest.exe tv all 2>&1 | tee -a "$TEST_RESULTS"
fi

# Release build, UBSan, C++11
if [ "$HAVE_UBSAN" -ne "0" ]; then
	echo "************************************" | tee -a "$TEST_RESULTS"
	echo "Release build, C++11, Undefined Behavior sanitizer" | tee -a "$TEST_RESULTS"
	echo
	
	unset CXXFLAGS
	make clean > /dev/null 2>&1
	export CXXFLAGS="-DNDEBUG -g2 -O2 -std=c++11 $ADD_CXXFLAGS"
	make ubsan
	./cryptest.exe v 2>&1 | tee -a "$TEST_RESULTS"
	./cryptest.exe tv all 2>&1 | tee -a "$TEST_RESULTS"
fi

# Release build, Asan, C++11
if [ "$HAVE_ASAN" -ne "0" ]; then
	echo "************************************" | tee -a "$TEST_RESULTS"
	echo "Release build, C++11, Address sanitizer" | tee -a "$TEST_RESULTS"
	echo
	
	unset CXXFLAGS
	make clean > /dev/null 2>&1
	export CXXFLAGS="-DNDEBUG -g2 -O2 -std=c++11 $ADD_CXXFLAGS"
	make asan
	./cryptest.exe v 2>&1 | tee -a "$TEST_RESULTS"
	./cryptest.exe tv all 2>&1 | tee -a "$TEST_RESULTS"
fi

# For Darwin, we need to test both -stdlib=libstdc++ (GNU) and
#  -stdlib=libc++ (LLVM) crossed with -std=c++03 and -std=c++11.
if [ "$IS_DARWIN" -ne "0" ]; then
	echo "************************************" | tee -a "$TEST_RESULTS"
	echo "Darwin, C++03, libc++ (LLVM)" | tee -a "$TEST_RESULTS"
	echo

	unset CXXFLAGS
	make clean > /dev/null 2>&1
	export CXXFLAGS="-DNDEBUG -g2 -O2 -std=c++03 -stdlib=libc++ $ADD_CXXFLAGS"
	make static cryptest.exe
	./cryptest.exe v 2>&1 | tee -a "$TEST_RESULTS"
	./cryptest.exe tv all 2>&1 | tee -a "$TEST_RESULTS"
fi

if [ "$IS_DARWIN" -ne "0" ]; then
	echo "************************************" | tee -a "$TEST_RESULTS"
	echo "Darwin, C++03, libstdc++ (GNU)" | tee -a "$TEST_RESULTS"
	echo

	unset CXXFLAGS
	make clean > /dev/null 2>&1
	export CXXFLAGS="-DNDEBUG -g2 -O2 -std=c++03 -stdlib=libstdc++ $ADD_CXXFLAGS"
	make static cryptest.exe
	./cryptest.exe v 2>&1 | tee -a "$TEST_RESULTS"
	./cryptest.exe tv all 2>&1 | tee -a "$TEST_RESULTS"
fi

if [ "$IS_DARWIN" -ne "0" ] && [ "$HAVE_CXX11" -ne "0" ]; then
	echo "************************************" | tee -a "$TEST_RESULTS"
	echo "Darwin, C++11, libc++ (LLVM)" | tee -a "$TEST_RESULTS"
	echo

	unset CXXFLAGS
	make clean > /dev/null 2>&1
	export CXXFLAGS="-DNDEBUG -g2 -O2 -std=c++11 -stdlib=libc++ $ADD_CXXFLAGS"
	make static cryptest.exe
	./cryptest.exe v 2>&1 | tee -a "$TEST_RESULTS"
	./cryptest.exe tv all 2>&1 | tee -a "$TEST_RESULTS"
fi

if [ "$IS_DARWIN" -ne "0" ] && [ "$HAVE_CXX11" -ne "0" ]; then
	echo "************************************" | tee -a "$TEST_RESULTS"
	echo "Darwin, C++11, libstdc++ (GNU)" | tee -a "$TEST_RESULTS"
	echo

	unset CXXFLAGS
	make clean > /dev/null 2>&1
	export CXXFLAGS="-DNDEBUG -g2 -O2 -std=c++11 -stdlib=libstdc++ $ADD_CXXFLAGS"
	make static cryptest.exe
	./cryptest.exe v 2>&1 | tee -a "$TEST_RESULTS"
	./cryptest.exe tv all 2>&1 | tee -a "$TEST_RESULTS"
fi

# Valgrind, C++03. Requires -O1 for accurate results
if [ "$HAVE_VALGRIND" -ne "0" ]; then
	echo "************************************" | tee -a "$TEST_RESULTS"
	echo "Valgrind, C++03" | tee -a "$TEST_RESULTS"
	echo
	
	unset CXXFLAGS
	make clean > /dev/null 2>&1
	export CXXFLAGS="-DNDEBUG -std=c++03 -g3 -O1 $ADD_CXXFLAGS"
	make static cryptest.exe
	valgrind ./cryptest.exe v 2>&1 | tee -a "$TEST_RESULTS"
	valgrind ./cryptest.exe tv all 2>&1 | tee -a "$TEST_RESULTS"
fi

# Valgrind, C++11. Requires -O1 for accurate results
if [ "$HAVE_VALGRIND" -ne "0" ] && [ "$HAVE_CXX11" -ne "0" ]; then
	echo "************************************" | tee -a "$TEST_RESULTS"
	echo "Valgrind, C++11" | tee -a "$TEST_RESULTS"
	echo
	
	unset CXXFLAGS
	make clean > /dev/null 2>&1
	export CXXFLAGS="-DNDEBUG -std=c++11 -g3 -O1 $ADD_CXXFLAGS"
	make static cryptest.exe
	valgrind ./cryptest.exe v 2>&1 | tee -a "$TEST_RESULTS"
	valgrind ./cryptest.exe tv all 2>&1 | tee -a "$TEST_RESULTS"
fi

TEST_END=$(date)

echo "************************************************" | tee -a "$TEST_RESULTS"
echo "************************************************" | tee -a "$TEST_RESULTS"
echo "Test started: $TEST_BEGIN" | tee -a "$TEST_RESULTS"
echo "Test finished: $TEST_END" | tee -a "$TEST_RESULTS"
echo | tee -a "$TEST_RESULTS"

COUNT=$(egrep '(error|FAILED)' cryptest-result.txt | wc -l)
if [ "$COUNT" -eq "0" ]; then
	echo "No failures detected" | tee -a "$TEST_RESULTS"
else
	echo "$COUNT errors detected" | tee -a "$TEST_RESULTS"
	egrep -n "(error|FAILED)" cryptest-result.txt
fi
echo "************************************************" | tee -a "$TEST_RESULTS"
echo "************************************************" | tee -a "$TEST_RESULTS"
